import base64
import hashlib
from Cryptodome.Cipher import AES
from Cryptodome.Random import get_random_bytes

def encrypt(raw, key):
    __key__ = hashlib.sha256(key).digest()
    BS = AES.block_size
    pad = lambda s: s + (BS - len(s) % BS) * chr(BS - len(s) % BS)

    raw = base64.b64encode(pad(raw).encode('utf8'))
    iv = get_random_bytes(AES.block_size)
    cipher = AES.new(key= __key__, mode= AES.MODE_CFB,iv= iv)
    return base64.b64encode(iv + cipher.encrypt(raw))

def decrypt(enc, key):
    __key__ = hashlib.sha256(key).digest()
    unpad = lambda s: s[:-ord(s[-1:])]

    enc = base64.b64decode(enc)
    iv = enc[:AES.block_size]
    cipher = AES.new(__key__, AES.MODE_CFB, iv)
    return unpad(base64.b64decode(cipher.decrypt(enc[AES.block_size:])).decode('utf8'))


from PIL import Image
from tools.translation_tools import text_to_binary

def modify_pixel(pix: int, bit: str) -> int:
    if bit == "1" and pix % 2 == 0:
            if pix < 255: pix += 1
            else: pix -= 1
    elif bit == "0" and pix % 2 != 0:
            if pix < 255: pix += 1
            else: pix -= 1
    
    return pix

def new_img_name(img_path: str) -> str:
    if "/" in img_path:
        img_path = img_path.split('/')
        filename = img_path[len(img_path)-1].split('.')
    elif "\\" in img_path:
        img_path = img_path.split('\\')
        filename = img_path[len(img_path)-1].split('.')
    else:
        filename = img_path.split('.')
    
    filename[0] += '-encoded'
    return '.'.join(filename)

def encode_img(img_path: str, text:str) -> None:
    """encode a text within an image (steganography)

    Args:
        img_path (str): the path of the image you want to encode
        text (str): the message you want to encode
    """
    text = f'{text} '
    img = Image.open(img_path)
    width, height = img.size
    binary_text = text_to_binary(text)
    len_data = len(binary_text)
    new_file_name = new_img_name(img_path)
    row = 0
    j=0

    for idx, byte in enumerate(binary_text):
        j+=1

        if j*3+1 >= width:
            row += 1
            j=0

        pix = list(img.getpixel((j*3+1,row)))
        for i in range(0, 3):
            pix[i] = modify_pixel(pix[i], byte[i])
        img.putpixel((j*3+1,row), tuple(pix))

        if j*3+2 >= width:
            row += 1
            j=0

        pix = list(img.getpixel((j*3+2,row)))
        for i in range(0, 3):
            pix[i] = modify_pixel(pix[i], byte[i+3])
        img.putpixel((j*3+2,row), tuple(pix))

        if j*3+3 >= width:
            row += 1
            j=0

        pix = list(img.getpixel((j*3+3,row)))
        for i in range(0, 2):
            pix[i] = modify_pixel(pix[i], byte[i+6])
        
        if len_data == j+1:
            pix[2] = modify_pixel(pix[2], "1")
        elif len_data > j+1:
            pix[2] = modify_pixel(pix[2], "0")
        
        img.putpixel((j*3+3,row), tuple(pix))
    
    img.save(new_file_name)


from PIL import Image
from tools.translation_tools import binary_to_text

def decode_pix(pix:int) -> str:
    if pix % 2 == 0: return "0"
    else: return "1"

def decode_img(img_path: str) -> str:
    """decode an image that has been encoded with this program

    Args:
        img_path (str): the path of the image you want to decode

    Returns:
        str: The decoded message of the image
    """
    img = Image.open(img_path)
    width, height = img.size
    binary_message = []
    finished_message, i = False, 0
    row = 0

    while finished_message == False:
        byte = ""
        i += 1

        for z in range(1, 3):
            if i*3+z >= width:
                row += 1
                i=0
            pix = list(img.getpixel((i*3+z, row)))
            for j in range(0, 3):
                byte += decode_pix(pix[j])

        if i*3+3 >= width:
                row += 1
                i=0
        pix = list(img.getpixel((i*3+3, row)))
        for j in range(0, 2):
            byte += decode_pix(pix[j])

        if decode_pix(pix[2]) == "1":
            finished_message = True

        binary_message.append(byte)

    return binary_to_text(binary_message)



from tools.encoding import encode_img
from tools.decoding import decode_img
from tools.encryption_tools import encrypt, decrypt
from tools.colors import colors
from bullet import Password
import os

title = colors.fg.green + """
  _____        _____ _                   
 |  __ \      / ____| |                  
 | |__) |   _| (___ | |_ ___  __ _  ___  
 |  ___/ | | |\___ \| __/ _ \/ _` |/ _ \ 
 | |   | |_| |____) | ||  __/ (_| | (_) |
 |_|    \__, |_____/ \__\___|\__, |\___/ 
         __/ |                __/ |      
        |___/                |___/       """ + colors.reset

def interface_tool():
    os.system('cls' if os.name == 'nt' else 'clear')
    print(title)
    print(colors.fg.darkgrey, "\nHide encrypted messages in your photos\nMade by Miquel Muntaner\nGithub repository: https://github.com/MiquelMuntaner/pystego\n", colors.reset)
    print("1. encrypt\n2. decrypt")
    option = input("Choose an option (1 or 2): ")

    if option == "1":
        path = input("\nWrite the path of the image you want to encrypt (.png): ")
        message = input("Write the message you want to encrypt: ")

        print(colors.fg.darkgrey, "\nYou can encrypt your message with a password so that not everyone can see the content using this program\nThis encryption is done with the very secure algorithm AES-256 bits", colors.reset)
        want_password = input("Do you want to use password?(y/n): ")

        if want_password == "y":
            print(colors.fg.darkgrey, "\nRemember that if you lose the password you will lose the message", colors.reset)
            password_match = False
            while password_match == False:
                password = Password(prompt="Type your password: ", hidden="*").launch()
                password2 = Password(prompt="Re-type your password: ", hidden="*").launch()
                password_match = password == password2

                if not password_match: print(colors.fg.red, "\nPasswords don't match", colors.reset)

            message = encrypt(message, password.encode())
        
        encode_img(path, str(message))
        print(colors.fg.green, colors.bold, "\nThe image has been encrypted", colors.reset)
    elif option == "2":
        path = input("\nWrite the path of the image you want to decode (.png): ")
        password = Password(prompt="Type the password (leave empty if there is no password): ", hidden="*").launch()

        message = decode_img(path)

        if len(password) != 0:
            message = decrypt(message.replace('b\'', '')[:-1].encode(), password.encode())

        print(colors.fg.green, colors.bold, "\nThe message has been decoded:", colors.reset)
        print(message)


from setuptools import setup, find_packages

setup(
    name='pystego',
    version='0.0.21',
    description='A Python program that allows you to save secret messages in images',
    author='Miquel Muntaner',
    author_email='miquemuntanerweb@gmail.com',
    url='https://github.com/MiquelMuntaner/pystego',
    py_modules=['pystego'],
    package_dir={'': 'src'},
    packages=find_packages(include=('colors', 'decoding', 'encoding', 'encryption_tools', 'interface', 'translation_tools')),
    install_requires=['bullet', 'click', 'Pillow', 'pycryptodomex', 'typer'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License"
    ]
)


def text_to_binary(text:str) -> list:
    """translates text to binary

    Args:
        text (str): the text you want to translate

    Returns:
        list: the text translated to binary in a list with each byte
    """
    binary = [bin(char).replace("0b", "") for char in [ord(char) for char in text]]
    return ["0"*(8-len(byte)) + byte for byte in binary if len(byte) < 8]

def binary_to_text(binary:list) -> str:
    """translates binary to text

    Args:
        binary (list): a list with each byte in a string

    Returns:
        str: a string with the content already translated
    """
    return "".join([chr(int(byte, 2)) for byte in binary])


from tools.encoding import encode_img
from tools.decoding import decode_img
from tools.encryption_tools import encrypt, decrypt
from tools.colors import colors
from tools.interface import interface_tool
from bullet import Password
from pathlib import Path
import os
import typer

app = typer.Typer()

@app.command()
def interface():
    """Use an interface in the terminal to encode and decode images more easily"""
    interface_tool()

@app.command()
def encode(message: str, img_path: Path, encryption: bool = typer.Option(False, "--encrypt", "-e", help="Secure encryption with AES-256 bit algorithm")):
    """Encode a message in an image, you can use AES-256 bit encryption if you use the -e flag"""

    if encryption == False:
        print(colors.fg.yellow, "Warning: you are not using encryption anyone with this program will be able to see the message. To use encryption type -e at the end of the command.", colors.reset)
    else:
        print(colors.fg.darkgrey, "\nRemember that if you lose the password you will lose the message", colors.reset)
        password_match = False
        while password_match == False:
            password = Password(prompt="Type your password: ", hidden="*").launch()
            password2 = Password(prompt="Re-type your password: ", hidden="*").launch()
            password_match = password == password2

            if not password_match: print(colors.fg.red, "\nPasswords don't match", colors.reset)

        message = encrypt(message, password.encode())
    
    encode_img(str(img_path), str(message))
    print(colors.fg.green, colors.bold, "\nThe image has been encrypted", colors.reset)

@app.command()
def decode(img_path: Path, encryption: bool = typer.Option(False, "--encrypted", "-e", help="Secure encryption with AES-256 bit algorithm")):
    message = decode_img(img_path)
    if encryption:
        password = Password(prompt="Type the password (leave empty if there is no password): ", hidden="*").launch()
        message = decrypt(message.replace('b\'', '')[:-1].encode(), password.encode())
        
    print(colors.fg.green, colors.bold, "\nThe message has been decoded:", colors.reset)
    print(message)

if __name__ == "__main__":
    app()